;; Direct calculation of covering radius for a code of length NC

(proclaim '(type fixnum NC CZ RZ R))

(defvar NC   13)    ;; Length of column
(defvar CZ 8192)    ;; Size of space (= vp-set) = 2^NC
(defvar RZ    4)    ;; No. of bits to represent covering radius
(defvar R     0)    ;; Radius

(*proclaim '(type Boolean-pvar H S X Y Z))

(*defvar H nil!!)    ;; The generating set
(*defvar S nil!!)    ;; Set covered so far
(*defvar X nil!!)    ;; Set just covered
(*defvar Y nil!!)    ;; Working set

(*proclaim '(type (field-pvar NC) V W))

(*defvar V (!! 0))   ;; Vector (= self-address)
(*defvar W (!! 0))   ;; Working vector

(*proclaim '(type (field-pvar RZ) VR))

(*defvar VR (!! 0))  ;; Radius at which V covered

;; Setup for code of sz bits

(defun setup (sz)
  (setq NC sz)
  (setq CZ (ash 1 sz))
  (setq RZ (integer-length sz))
  (*cold-boot :initial-dimensions (list CZ))
)

;; Initialize generators consisting of binary reps of 2^i

(defun test-basis ()
  (*set H nil!!)
  (dotimes (i NC) (*setf (pref H (ash 1 i)) T))
)

(defun random-gen (r)
  (*set H (<!! (random!! (!! 100000)) (!! r)))
  (*setf (pref H 0) nil)
  (list (*when H (*sum (!! 1))) 'VECTORS)
)

;; Calculate covering radius - H should already be defined

(defun cover ()
  (*set V (self-address!!))         ;; Vector elements
  (*set S H)  (*setf (pref S 0) T)  ;; Set covered initially
  (*set X H)
  (*set VR (!! 0))
  (setq R 1)                       ;; Initial radius
  (*when H (*set VR (!! R)))
  ;; The PROG implements a do-while-repeat loop
  (PROG NIL
    LP  (format t "R = ~D,  X has ~D elements~%"
                        R   (*when X (*sum (!! 1))))
        (if (*and S) (go EX))  ;; Finished if S = t!! (all covered)
        ;; Y := H + X
        (*set Y nil!!)
        (*when H (do-for-selected-processors (i)
          (*all (*when X
            (*set W (logxor!! (!! i) V))
            (*pset :or t!! Y W))) ; end *when X, end *all
        )) ; end *when H
        ;; X := Y \ S
        (*set X (and!! Y (not!! S)))
        ;; S := S union Y
        (*set S (or!! S Y))
	(incf R)                    ;; Next radius
	(*when X (*set VR (!! R)))
        (go LP)                     ;; Repeat
    EX  ;; Loop finished
  ) ;; end prog
  R ;; returns radius
)

;; Install "cyclic" matrix for specifying H.  "file" is a string
;; naming a file containing the first row, padded with leading 0's.
;; Values are read as integers (0/1).
;; File should begin with an integer specifying the number of entries.

(*proclaim '(type Boolean-pvar MX))

(*defvar MX nil!!)    ; Represents matrix as a set of processors

(*proclaim '(type (field-pvar 1) MR))

(*defvar MR (!! 0))    ; MX is one row of the matrix

(defun ldcyclic (file)
  (with-open-file (instr file)
    (let ((nbits (read instr)))  ;; Number of bits in vector
      ;; Read initial vector
      (*set MR (!! 0))
      (dotimes (i nbits)
	(*setf (pref MR (grid i)) (read instr)))
      ;; Construct matrix in W
      (*set W (!! 0))
      (dotimes (i NC)
        (*set W (dpb!! MR (byte!! (!! 1) (!! (- NC i 1))) W))
	(*news MR MR -1))
      ;; Construct MX
      (*set MX nil!!)
      (*when (<!! (self-address-grid!! (!! 0)) (!! nbits))
	 (*pset :or t!! MX W))
    ))  ; end let, end with-open-file
)

;; Set H to subset of MX obtained by excluding sz vectors beginning at bg,
;; and proceeding in steps of stp, wraps around if appropriate, but in
;; that case may not actually exclude sz vectors.

(*proclaim '(type (signed-pvar (+ NC 3)) DL))

(*defvar DL (!! 0))    ; For subset calculations

(defun msubset (bg sz stp)
  (*set H MX)
  (*when MX
    (let* ((xsz (*sum (!! 1)))                   ; Number of vectors in MX
	   (mmx (truncate (* (1- sz) stp) xsz))) ; Max value for m
      (*set DL (-!! (enumerate!!) (!! bg)))
      (*when (<!! DL (!! 0)) (*set DL (+!! DL (!! xsz))))
      (dotimes (m mmx)
	(*if (plusp!! (rem!! DL (!! stp))) (*set DL (+!! DL (!! xsz)))))
      (*when (and!! (zerop!! (rem!! DL (!! stp)))
		    (<!! DL (!! (* stp sz))))
	 (*set H nil!!))))
)

;; Print Boolean pvar as 0/1 array, 64 entries/line

(*defun ppb (bpv)
  (dotimes (i *number-of-processors-limit*)
    (if (zerop (rem i 64)) (terpri))
    (write-char (if (pref bpv i) #\1 #\0)))
)